#define GLFW_INCLUDE_NONE
#pragma once
#include <GLFW/glfw3.h>
#include <chrono>
#include <glm/glm.hpp>
#include <imgui.h>
#include <imgui_impl_glfw.h>
#include <imgui_impl_opengl3.h>
#include <iostream>
#include <thread>
#include "kernel.cuh"


#define NUM_VERTICES 100
//implement unoptimised version of boids algorithm https://en.wikipedia.org/wiki/Boids

const char *vertexShaderSrc = R"(
#version 450

layout (location = 0) in vec2 aPos;

void main() {
    gl_Position = vec4(aPos, 1, 1);
}

)"; 

const char *fragmentShaderSrc = R"(
#version 450

layout (location = 0) out vec4 outColor;

void main() {
    outColor = vec4(1, 0, 0, 1);
}

)";

void message_callback(GLenum source, GLenum type, GLuint id, GLenum severity,
                      GLsizei length, GLchar const *message,
                      void const *user_param) {
  if (severity == GL_DEBUG_SEVERITY_NOTIFICATION)
    return;

  auto const src_str = [source]() {
    switch (source) {
    case GL_DEBUG_SOURCE_API:
      return "API";
    case GL_DEBUG_SOURCE_WINDOW_SYSTEM:
      return "WINDOW SYSTEM";
    case GL_DEBUG_SOURCE_SHADER_COMPILER:
      return "SHADER COMPILER";
    case GL_DEBUG_SOURCE_THIRD_PARTY:
      return "THIRD PARTY";
    case GL_DEBUG_SOURCE_APPLICATION:
      return "APPLICATION";
    case GL_DEBUG_SOURCE_OTHER:
      return "OTHER";
    default:
      return "UNKNOWN SOURCE";
    }
  }();

  auto const type_str = [type]() {
    switch (type) {
    case GL_DEBUG_TYPE_ERROR:
      return "ERROR";
    case GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR:
      return "DEPRECATED_BEHAVIOR";
    case GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR:
      return "UNDEFINED_BEHAVIOR";
    case GL_DEBUG_TYPE_PORTABILITY:
      return "PORTABILITY";
    case GL_DEBUG_TYPE_PERFORMANCE:
      return "PERFORMANCE";
    case GL_DEBUG_TYPE_MARKER:
      return "MARKER";
    case GL_DEBUG_TYPE_OTHER:
      return "OTHER";
    default:
      return "UNKNOWN TYPE";
    }
  }();

  auto const severity_str = [severity]() {
    switch (severity) {
    case GL_DEBUG_SEVERITY_NOTIFICATION:
      return "NOTIFICATION";
    case GL_DEBUG_SEVERITY_LOW:
      return "LOW";
    case GL_DEBUG_SEVERITY_MEDIUM:
      return "MEDIUM";
    case GL_DEBUG_SEVERITY_HIGH:
      return "HIGH";
    default:
      return "UNKNOWN SEVERITY";
    }
  }();
  std::cout << src_str << ", " << type_str << ", " << severity_str << ", " << id
            << ": " << message << '\n';
}

int main(int argc, char *argv[]) {
    //int N = 1024;
    //int* a, * b, * c;
    //int size = N * sizeof(int);

    //// Allocate memory on the GPU
    //cudaMalloc(&a, size);
    //cudaMalloc(&b, size);
    //cudaMalloc(&c, size);

    //// Initialize input arrays
    //int* h_a = new int[N];
    //int* h_b = new int[N];
    //for (int i = 0; i < N; i++) {
    //    h_a[i] = i*2;
    //    h_b[i] = N - i;
    //}

    //// Copy input arrays to GPU memory
    //cudaMemcpy(a, h_a, size, cudaMemcpyHostToDevice);
    //cudaMemcpy(b, h_b, size, cudaMemcpyHostToDevice);

    //// Launch the CUDA kernel
    //int threadsPerBlock = 256;
    //int blocksPerGrid = (N + threadsPerBlock - 1) / threadsPerBlock;
    //addKernelLaunch(blocksPerGrid, threadsPerBlock, a, b, c);

    //// Copy the result back to the host
    //int* h_c = new int[N];
    //cudaMemcpy(h_c, c, size, cudaMemcpyDeviceToHost);

    //// Print the result
    //for (int i = 0; i < N; i++) {
    //    std::cout << h_c[i] << " ";
    //}
    //std::cout << std::endl;

    //// Free memory
    //delete[] h_a;
    //delete[] h_b;
    //delete[] h_c;
    //cudaFree(a);
    //cudaFree(b);
    //cudaFree(c);

  glfwInit();
  glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 4);
  glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 5);
  glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);

  auto window = glfwCreateWindow(1000, 1000, "Example", nullptr, nullptr);
  if (!window)
    throw std::runtime_error("Error creating glfw window");
  glfwMakeContextCurrent(window);
  glfwSwapInterval(1);

  if (!gladLoaderLoadGL())
    throw std::runtime_error("Error initializing glad");

  /**
   * Initialize ImGui
   */
  IMGUI_CHECKVERSION();
  ImGui::CreateContext();
  ImGui_ImplGlfw_InitForOpenGL(window, true);
  ImGui_ImplOpenGL3_Init("#version 450 core");
  ImGui::StyleColorsClassic();

  glDisable(GL_CULL_FACE);
  glEnable(GL_DEBUG_OUTPUT);
  glDebugMessageCallback(message_callback, nullptr);

  /**
   * Compile shader
   */
  int success;
  char infoLog[512];
  auto vertexShader = glCreateShader(GL_VERTEX_SHADER);
  glShaderSource(vertexShader, 1, &vertexShaderSrc, 0);
  glCompileShader(vertexShader);

  glGetShaderiv(vertexShader, GL_COMPILE_STATUS, &success);
  if (!success) {
    glGetShaderInfoLog(vertexShader, 512, nullptr, infoLog);
    std::cerr << "Vertex shader compilation failed:" << std::endl;
    std::cerr << infoLog << std::endl;
    return 0;
  }

  auto fragmentShader = glCreateShader(GL_FRAGMENT_SHADER);
  glShaderSource(fragmentShader, 1, &fragmentShaderSrc, 0);
  glCompileShader(fragmentShader);

  glGetShaderiv(fragmentShader, GL_COMPILE_STATUS, &success);
  if (!success) {
    glGetShaderInfoLog(fragmentShader, 512, nullptr, infoLog);
    std::cerr << "Fragment shader compilation failed:" << std::endl;
    std::cerr << infoLog << std::endl;
    return 0;
  }

  auto program = glCreateProgram();
  glAttachShader(program, vertexShader);
  glAttachShader(program, fragmentShader);
  glLinkProgram(program);

  glGetProgramiv(program, GL_LINK_STATUS, &success);
  if (!success) {
    glGetShaderInfoLog(program, 512, nullptr, infoLog);
    std::cerr << "Shader linking failed:" << std::endl;
    std::cerr << infoLog << std::endl;
    return 0;
  }

  glDetachShader(program, vertexShader);
  glDetachShader(program, fragmentShader);

  /**
   * Create vertex array and buffers
   */
  GLuint VBO, VAO;
  int elementSize = 2;
  float vertexArray[NUM_VERTICES*2];
  for (int i = 0; i < NUM_VERTICES; i++) {
      vertexArray[i * 2] = i / (0.5*NUM_VERTICES)-1;
      vertexArray[i * 2+1] = i / (0.5*NUM_VERTICES) -1;
  }
  glGenVertexArrays(1, &VAO);
  glGenBuffers(1, &VBO);


  // bind the Vertex Array Object first, then bind and set vertex buffer(s), and then configure vertex attributes(s).
  glBindVertexArray(VAO);

  glBindBuffer(GL_ARRAY_BUFFER, VBO);
  glBufferData(GL_ARRAY_BUFFER, NUM_VERTICES * elementSize * sizeof(float), vertexArray, GL_DYNAMIC_DRAW);

  glVertexAttribPointer(0, elementSize, GL_FLOAT, GL_FALSE, elementSize * sizeof(float), (void*)0);
  glEnableVertexAttribArray(0);

  // note that this is allowed, the call to glVertexAttribPointer registered VBO as the vertex attribute's bound vertex buffer object so afterwards we can safely unbind
  glBindBuffer(GL_ARRAY_BUFFER, 0);

  cudaGraphicsResource* vbo_resource;
  createVBO(VBO, &vbo_resource);
  // You can unbind the VAO afterwards so other VAO calls won't accidentally modify this VAO, but this rarely happens. Modifying other
  // VAOs requires a call to glBindVertexArray anyways so we generally don't unbind VAOs (nor VBOs) when it's not directly necessary.
  glBindVertexArray(0);
  glUseProgram(program);
  glClearColor(1, 1, 1, 1);

  float speed = 0.01;
  while (!glfwWindowShouldClose(window)) {
    runCuda(&vbo_resource, NUM_VERTICES,speed);
    glfwPollEvents();
    glClear(GL_COLOR_BUFFER_BIT);
    glPointSize(3);
    glBindVertexArray(VAO);
    glDrawArrays(GL_POINTS, 0, NUM_VERTICES);
    glBindVertexArray(0);

    ImGui_ImplOpenGL3_NewFrame();
    ImGui_ImplGlfw_NewFrame();
    ImGui::NewFrame();

    static bool showDemo = false;
    ImGui::Begin("Example");
    ImGui::SliderFloat("speed", &speed, 0, 0.1,"%.4f");
    ImGui::End();

    ImGui::Render();
    ImGui_ImplOpenGL3_RenderDrawData(ImGui::GetDrawData());

    glfwSwapBuffers(window);

    //runCuda(&cuda_vbo_resource);
    std::this_thread::sleep_for(std::chrono::milliseconds(1));
  }

  ImGui_ImplOpenGL3_Shutdown();
  ImGui_ImplGlfw_Shutdown();
  ImGui::DestroyContext();

  glfwDestroyWindow(window);
  glfwTerminate();
  return 0;
}
